# =====================================================================
# xmllib.awk: XML AWK function library for NoSQL.
#
# Copyright (c) 2007 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 dated June, 1991.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# =====================================================================

# =====================================================================
# void __rdf([string root, string child, string about,\
#	    string struct, string property, string name, string value,\
#	    array aname, array avalue, boolean noauto,boolean noenc])
#
# Experimental table-to-rdf converter. The calling program is required
# to have a knowledge of the input table structure and semantic, and
# it will have to call this function in the proper way.
#
# Here are some tips on how the underlying logic works, as it may look
# rather obscure ad a first glance. This program has proven flexible
# enough to handle the creation of rather complex and nested structures,
# like http://linuxgazette.net/105/misc/oregan/foaf-example.rdf.txt
#
# Here are the rules adopted, with respect to the above example:
#
# <foaf:Person> is a "root" element.
# <foaf:name>Jimmy O'Regan</foaf:name> is a name/value pair.
# <foaf:based_near> is a "child" element.
# <foaf:Point> is a "property" element.
# <geo:lat>52.6796</geo:lat> is a name/value pair.
#
# * Property elements can only appare inside RDF structs (Seq/Alt/Bag).
# * Root and property elements may take an "about" attribute, while 
#   child and name elements may not.
# * Property elements may only appare inside child nodes, which implies
#   that child nodes require an RDF struct.
# * Attributes are accepted only for name elements, and only if the
#   associated value is an URI.
# * Root elements may have name/value pairs, but not property elements
#   neither RDF structs, unless these are incapsulated inside child
#   elements.
# * A child element may not have name/value pairs, unless these are
#   incapsulated inside a struct of properties.
#
# If the above rules are violated, the program will ignore the failing
# arguments and apply sensibly the correct ones, in order for the output
# to parse as valid XML/RDF. So if you wander why the output does not
# look like what you expected, either there's a bug in this program or
# you have got these rules wrong.
#
# This program looks flexible enough to allow for the the generation of
# arbitrarily complex XML structures, although I have not yet tested it
# extensively in that respect.
# =====================================================================

function __rdf(root,child,about,struct,property,name,value,aname,\
		      avalue,noauto,noenc,			  a) {

  # make sure file is appended to, or printing multiple RDF
  # response blocks with different ontologies will not be possible!

  printf(_NULL) >> __STDOUT

  if (__isuri(tolower(about)) == _TRUE || about ~ /^#./)
		about = " rdf:about=\"" __xmlencode(about) "\""
  else if (about == ".") about = " rdf:about=\"\""
  else about = ""

  if (struct == "Alt") struct = "rdf:Alt"
  else if (struct == "Bag") struct = "rdf:Bag"
  else struct = "rdf:Seq"				# default.

  if (root != "") {					# new root ?

     # close previous tags if necessary.

     if (__XML["property"] != "")
	printf("</%s>\n</rdf:li>\n",__XML["property"]) > __STDOUT

     if (__XML["struct"] != "") printf("</%s>\n",__XML["struct"]) > __STDOUT

     if (__XML["child"] != "") printf("</%s>\n",__XML["child"]) > __STDOUT

     if (__XML["root"] != "") printf("</%s>\n",__XML["root"]) > __STDOUT

     # clear old values.
     __XML["child"] = __XML["struct"] = __XML["property"] = ""

     # the fake root value "." simply closes the RDF output.

     if (root == ".") {
	close(file)
	__XML["root"] = _NULL		# this may be tested elsewhere.
	return
     }

     # set new root.
     printf("<%s%s>\n",root,about) > __STDOUT; __XML["root"] = root
  }

  if (child != "") {					# new child ?

     # close previous tags if necessary.

     if (__XML["property"] != "")
	printf("</%s>\n</rdf:li>\n",__XML["property"]) > __STDOUT

     if (__XML["struct"] != "") printf("</%s>\n",__XML["struct"]) > __STDOUT

     if (__XML["child"] != "") printf("</%s>\n",__XML["child"]) > __STDOUT

     # clear old values.
     __XML["struct"] = __XML["property"] = ""

     # the fake child value "." simply closes the current child.
     if (child == ".") return

     # set new child.
     printf("<%s>\n<%s>\n",child,struct) > __STDOUT

     __XML["child"] = child; __XML["struct"] = struct
  }

  if (property != "" && __XML["child"] != "") {		# new property ?

     # Properties can only occur inside childs (use default namespace).
     if (__XML["child"] == "") __rdf("","csa:Child")

     # close previous tags if necessary.

     if (__XML["property"] != "")
		printf("</%s>\n</rdf:li>\n",__XML["property"]) > __STDOUT

     # the fake property value "." simply closes the current property.
     if (property == ".") return

     # set new property.

     printf("<rdf:li>\n<%s%s>\n",property,about) > __STDOUT

     __XML["property"] = property
  }

  if (name != "") {

     if (!noauto && __isuri(tolower(value)) == _TRUE) {

	printf("<%s",name) > __STDOUT

	# attributes are taken into consideration only for URI-type values.
	for (a in aname)
	    printf(" %s=\"%s\"",aname[a],\
			__xmlencode(avalue[aname[a]])) > __STDOUT

	printf(" rdf:resource=\"%s\" />\n",__xmlencode(value)) > __STDOUT
     }

     else {
	if (noenc) printf("<%s>%s</%s>\n",name,value,name) > __STDOUT
	else printf("<%s>%s</%s>\n",name,__xmlencode(value),name) > __STDOUT
     }
  }

  # now that the array contents have been consumed I clear them,
  # as otherwise they would need to be cleared by the calling
  # program, or they could be attached also to other values they
  # do not apply to.

  delete aname; delete avalue
}

# =====================================================================
# boolean __isuri(string uri)
#
# Return _TRUE if the received string is a sintactically valid URI, or
# _FALSE otherwise.
# =====================================================================

function __isuri(uri) {

  # This pattern matches URI in a very broad manner, but on the other
  # hand it is rather difficult to be more selective. Furthermore, it
  # does not currently check whether the received string contains
  # invalid URI characters, beside SPACE, '\t', '\r', and '\n'.

  if (uri ~ /^[a-z][a-z0-9]+:[^ \t\r\n]+$/) return _TRUE
  else return _FALSE
}

# =====================================================================
# string __xmlencode(string s)
#
# Encode special XML characters into the equivalent ISO entities. The
# set of characters that are XML-encoded is:  & < " \001
#
# Warning: '&' must be escaped first. According to the XML
# specification, characters 0 through 31 and character 127 may not
# appear anywhere in a document, not even as character references, such
# as &#31;.
# =====================================================================

function __xmlencode(string) {

   gsub(/&/, "\\&amp;", string)			# Ampersand at first.
   gsub(/</, "\\&lt;", string)			# Open tag
   gsub(/"/, "\\&quot;", string)		# double quote
   gsub(/'/, "\\&#39;", string)

   return string
}

# =====================================================================
# End of xmllib.awk
# =====================================================================
